# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models
from django.db.models import Q
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator
from django.utils.translation import ugettext_lazy as _
from netfields import CidrAddressField, NetManager
from netaddr import IPNetwork, IPSet


def validate_subnet(cidr):
    """Checks that a CIDR object is indeed a subnet, i.e. the host bits are
    all set to zero."""
    from django.utils.translation import ugettext as _
    if not isinstance(cidr, IPNetwork):
        raise ValidationError(_("Error: expected IPNetwork object."))
    if cidr.ip != cidr.network:
        error = _("{subnet} is not a proper subnet, did you mean {correct}?")
        raise ValidationError(error.format(subnet=cidr, correct=cidr.cidr))


class IPPool(models.Model):
    """Pool of IP addresses (either v4 or v6)."""
    name = models.CharField(max_length=255, blank=False, null=False,
                            verbose_name=_('name'),
                            help_text=_("Name of the IP pool"))
    default_subnetsize = models.PositiveSmallIntegerField(
        blank=False,
        verbose_name=_("default subnet size"),
        help_text=_("Default size of subnets allocated to subscribers "
                    "from this pool"),
        validators=[MaxValueValidator(64)])
    inet = CidrAddressField(validators=[validate_subnet],
                            verbose_name=_("network"),
                            help_text=_("Address space to use"))
    objects = NetManager()

    def clean(self):
        from django.utils.translation import ugettext as _
        if self.inet:
            max_subnetsize = 64 if self.inet.version == 6 else 32
            if not self.inet.prefixlen <= self.default_subnetsize <= max_subnetsize:
                raise ValidationError(_("Invalid default subnet size."))
            # Check that related subnet are in the pool (useful when
            # modifying an existing pool that already has subnets
            # allocated in it)
            incorrect = [str(subnet) for subnet in self.ipsubnet_set.all()
                         if not subnet.inet in self.inet]
            if incorrect:
                err = _("Some subnets allocated in this pool are outside "
                        "the pool: {subnets}.")
                raise ValidationError(err.format(subnets=incorrect))

    def __unicode__(self):
        return self.name

    class Meta:
        verbose_name = _("IP pool")
        verbose_name_plural = _("IP pools")


class IPSubnet(models.Model):
    inet = CidrAddressField(blank=True, validators=[validate_subnet],
                            unique=True, verbose_name=_("subnet"),
                            help_text=_("Leave empty for automatic allocation"))
    objects = NetManager()
    ip_pool = models.ForeignKey(IPPool, verbose_name=_("IP pool"))
    configuration = models.ForeignKey('configuration.Configuration',
                                      related_name='ip_subnet',
                                      verbose_name=_("configuration"))
    delegate_reverse_dns = models.BooleanField(
        default=False,
        verbose_name=_("delegate reverse DNS"),
        help_text=_("Whever to delegate reverse DNS resolution for this "
                    "subnet to one or more name servers"))
    name_server = models.ManyToManyField(
        'reverse_dns.NameServer',
        blank=True,
        verbose_name=_("name server"),
        help_text=_("Name server to use for the delegation of reverse DNS"))

    def allocate(self):
        """Automatically allocate a free subnet"""
        from django.utils.translation import ugettext as _
        pool = IPSet([self.ip_pool.inet])
        used = IPSet((s.inet for s in self.ip_pool.ipsubnet_set.all()))
        free = pool.difference(used)
        # Generator for efficiency (we don't build the whole list)
        available = (p for p in free.iter_cidrs() if p.prefixlen <= self.ip_pool.default_subnetsize)
        # TODO: for IPv4, get rid of the network and broadcast
        # addresses? Not really needed nowadays, and we usually don't
        # have a real subnet in practice (i.e. Ethernet segment), but
        # many /32.
        try:
            first_free = available.next()
        except StopIteration:
            raise ValidationError(_("Unable to allocate an IP subnet in the "
                                    "specified pool: not enough space left."))
        # first_free is a subnet, but it might be too large for our needs.
        # This selects the first sub-subnet of the right size.
        self.inet = first_free.subnet(self.ip_pool.default_subnetsize, 1).next()

    def validate_inclusion(self):
        """Check that we are included in the IP pool"""
        from django.utils.translation import ugettext as _
        if not self.inet in self.ip_pool.inet:
            raise ValidationError(_("Subnet must be included in the IP pool."))
        # Check that we don't conflict with existing subnets.
        conflicting = self.ip_pool.ipsubnet_set.filter(Q(inet__net_contained_or_equal=self.inet) |
                                                       Q(inet__net_contains_or_equals=self.inet)).exclude(id=self.id)
        if conflicting:
            error = _("Subnet must not intersect with existing subnets.\n"
                      "Intersected subnets: {subnets}.")
            raise ValidationError(error.format(subnets=conflicting))

    def validate_reverse_dns(self):
        """Check that reverse DNS entries, if any, are included in the subnet"""
        from django.utils.translation import ugettext as _
        incorrect = [str(rev.ip) for rev in self.reversednsentry_set.all() if not rev.ip in self.inet]
        if incorrect:
            error = _("Some reverse DNS entries are not in the subnet: "
                      "{entries}.")
            raise ValidationError(error.format(entries=incorrect))

    def clean(self):
        if not self.inet:
            self.allocate()
        else:
            self.validate_inclusion()
        self.validate_reverse_dns()

    def __unicode__(self):
        return str(self.inet)

    class Meta:
        verbose_name = _("IP subnet")
        verbose_name_plural = _("IP subnets")
